#A part of NonVisual Desktop Access (NVDA)
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.
#Copyright (C) 2012 Ulf Beckmann <beckmann@flusoft.de>

# Parts of this code are inherited from the baum braille driver
# written by James Teh <jamie@jantrid.net>

# This file represents the braille display driver for
# Seika3 V2.00, Seika80, a product from Nippon Telesoft
# see www.seika-braille.com for more details
# 01.06.2012

import time
import wx
import serial
import braille
import inputCore
import hwPortUtils
from logHandler import log

TIMEOUT = 0.2
BAUDRATE = 9600
READ_INTERVAL = 50

CR_BEGIN = 0x100
CR_END = 0x200
CELL_COUNT80 = 80
CELL_COUNT40 = 40
CELL_COUNT16 = 16



SEIKA_QUERY  = chr(0xFF) + chr(0xFF) + chr(0x1C)
SEIKA_SENDHEADER80 = chr(0xff)+chr(0xff)+chr(0x73)+chr(0x38)+chr(0x30)+chr(0)+chr(0)+chr(0)
SEIKA_SENDHEADER40 = chr(0xff)+chr(0xff)+chr(0x73)+chr(0x65)+chr(0x69)+chr(0x6b)+chr(0x61)+chr(0)
SEIKA_SENDHEADER_COUNT = 8

SEIKA_SENDHEADER = SEIKA_SENDHEADER40

SEIKA_KEYMAP = [
	 "LEFT",
	 "B1", 
	 "B2",
	 "B3",
	 "B4",
	 "B5",
	 "B6",
	 "RIGHT"]


USB_IDS = frozenset((
	"USB\\Vid_10c4&Pid_ea60", # Seika USB to Ser
        "BTHENUM\\{00001101-0000-1000-8000-00805f9b34fb}"
#
))

class BrailleDisplayDriver(braille.BrailleDisplayDriver):
	name = "Seika"
	description = "Seika Braille"

	numCells = CELL_COUNT80
	@classmethod
	def check(cls):
		return True

	def __init__(self):
		super(BrailleDisplayDriver, self).__init__()

		for portInfo in hwPortUtils.listComPorts(onlyAvailable=True):
			port = portInfo["port"]
			hwID = portInfo["hardwareID"]
			# log.info("Found port {port} with hardwareID {hwID}".format(port=port, hwID=hwID))
			if not hwID.startswith(r"USB"):
				continue
			try:
				usbID = hwID.split("&R", 1)[0]
			except IndexError:
				continue
			if usbID not in USB_IDS:
				continue
			# At this point, a port bound to this display has been found.
			# Try talking to the display.
			try:
				self._ser = serial.Serial(port, baudrate=BAUDRATE, timeout=TIMEOUT, writeTimeout=TIMEOUT, parity=serial.PARITY_ODD, bytesize=serial.EIGHTBITS, stopbits=serial.STOPBITS_ONE)
			except serial.SerialException:
				continue
			cells = SEIKA_QUERY
			self._ser.write(cells)
			self._ser.flush()
			# Read out the input buffer
			time.sleep(0.1)
			versionS = self._ser.read(13)
			if versionS.startswith(r"seika80"):
				log.info("Found Seika80 connected via {port} Version {versionS}".format(port=port, versionS=versionS))
				self.numCells = CELL_COUNT80
				self.display([0])
				SEIKA_SENDHEADER = SEIKA_SENDHEADER80
				break
			elif versionS.startswith(r"seika3"):
				log.info("Found Seika40 connected via {port} Version {versionS}".format(port=port, versionS=versionS))
				self.numCells = CELL_COUNT40
				self.display([0])
				break
		else:
			raise RuntimeError("No SEIKA display found")
		self._readTimer = wx.PyTimer(self.handleResponses)
		self._readTimer.Start(READ_INTERVAL)

	def terminate(self):
		try:
			super(BrailleDisplayDriver, self).terminate()
			self._readTimer.Stop()
			self._readTimer = None
		finally:
			self._ser.close()

	def display(self, cells):
		# every transmitted line consists of the preamble SEIKA_SENDHEADERxx and the Cells
		if self.numCells==80:
			line = SEIKA_SENDHEADER80 + "".join(chr(cell) for cell in cells)
		else:
			line = SEIKA_SENDHEADER40 + "".join(chr(0)+chr(cell) for cell in cells)
		expectedLength = SEIKA_SENDHEADER_COUNT + 80
		line += chr(0) * (expectedLength - len(line))
		self._ser.write(line)

	def handleResponses(self):
		chars = []
		anz = 0
		key = 0
		keys= set()
		max = 22 # for 80 cells
		if self.numCells == 40:
			max = 12
		while self._ser.inWaiting() and anz<max:
			char = self._ser.read(1)
			if char:
				chars.append(ord(char))
				anz = anz + 1
				time.sleep(0.005)
		# if anz > 0:
		#	log.info("Seika K {anz}".format(anz=anz))
		if anz == 2: # normal key
			chars[0] = chars[0] & 0x1f
			chars[1] = chars[1] & 0x1a
			if chars[0] & 1: # LEFT
				keys.add(SEIKA_KEYMAP[0])
			if chars[0] & 4: # RIGHT
				keys.add(SEIKA_KEYMAP[7])
			if chars[1] & 2: # B1 
				keys.add(SEIKA_KEYMAP[1])
			if chars[1] & 8:
				keys.add(SEIKA_KEYMAP[2])
			if chars[1] & 16:
				keys.add(SEIKA_KEYMAP[3])
			if chars[0] & 16:
				keys.add(SEIKA_KEYMAP[4])
			if chars[0] & 8:
				keys.add(SEIKA_KEYMAP[5])
			if chars[0] & 2:
				keys.add(SEIKA_KEYMAP[6])
			data= "+".join(keys)
			# log.info("keys {keys}".format(keys=data))
			try:
				inputCore.manager.executeGesture(InputGestureKeys(data))
			except inputCore.NoInputGestureAction:
				log.debug("No Action for keys {keys}".format(keys=data))
				pass
		elif anz == 22 or anz == 12: # Routin 7 bytes offset + data
			i = 0
			k = 10
			if anz == 12: # 40 Cells 5*8
				k = 5
			while i < k:
				j = 0
				while j < 8:
					if chars[7+i] & (1<<j):
						key = i*8+j+1
					j = j + 1
				i = i + 1
			if key: # Routing key is pressed
				try:
					inputCore.manager.executeGesture(InputGestureRouting(key- 1))
				except inputCore.NoInputGestureAction:
					log.debug("No Action for routing command")
					pass

	gestureMap = inputCore.GlobalGestureMap({
		"globalCommands.GlobalCommands": {
			"braille_scrollBack": ("br(Seika):LEFT",),
			"braille_scrollForward": ("br(Seika):RIGHT",),
			"braille_previousLine": ("br(Seika):B3",),
			"braille_nextLine": ("br(Seika):B4",),
			"braille_toggleTether": ("br(Seika):B5",),
			"sayAll": ("br(Seika):B6",),
			"kb:tab": ("br(Seika):B1",),
			"kb:shift+tab": ("br(Seika):B2",),
			"kb:alt+tab": ("br(Seika):B1+B2",),
			"showGui": ("br(Seika):LEFT+RIGHT",),
			"braille_routeTo": ("br(Seika):routing",),
		},
	})

class InputGestureKeys(braille.BrailleDisplayGesture):

	source = BrailleDisplayDriver.name
	def __init__(self, keys):
		super(InputGestureKeys, self).__init__()
		self.id = keys

class InputGestureRouting(braille.BrailleDisplayGesture):

	source = BrailleDisplayDriver.name

	def __init__(self, index):
		super(InputGestureRouting, self).__init__()

		self.id = "routing"
		self.routingIndex = index
